/*************************************************************************\
 * Copyright (c) 2002 The University of Chicago, as Operator of Argonne
 * National Laboratory.
 * Copyright (c) 2002 The Regents of the University of California, as
 * Operator of Los Alamos National Laboratory.
 * This file is distributed subject to a Software License Agreement found
 * in the file LICENSE that is included with this distribution. 
\*************************************************************************/
/* program: cavget.c
 * purpose: vector CA gets from command line
 * M. Borland, 1995.
 */
#include "mdb.h"
#include "scan.h"
#include "match_string.h"
#include "SDDS.h"
#include "pvMultList.h"
#include <stdlib.h>
#include <time.h>
#include <cadef.h>
#include <epicsVersion.h>

#define CLO_LIST 0
#define CLO_RANGE 1
#define CLO_PENDIOTIME 2
#define CLO_DRYRUN 3
#define CLO_LABELED 4
#define CLO_FLOATFORMAT 5
#define CLO_DELIMITER 6
#define CLO_NOQUOTES 7
#define CLO_EMBRACE 8
#define CLO_REPEAT 9
#define CLO_ERRORVALUE 10
#define CLO_NUMERICAL 11
#define CLO_CAVPUTFORM 12
#define CLO_EXCLUDEERRORS 13
#define CLO_STATISTICS 14
#define CLO_DESPIKE 15
#define CLO_PRINTERRORS 16
#define COMMANDLINE_OPTIONS 17
char *commandline_option[COMMANDLINE_OPTIONS] = {
  "list", "range", "pendiotime", "dryrun", "labeled",
  "floatformat", "delimiter", "noquotes", "embrace", "repeat",
  "errorvalue", "numerical", "cavputform", "excludeerrors","statistics","despike", "printErrors",
} ;

#define CLO_STATS_TAGVALUE 0
#define CLO_STATS_PRETTY 1
#define CLO_STATS_SDDS 2
#define STATISICS_FORMATS 3
char *statisticsFormat[STATISICS_FORMATS]={
  "tagvalue","pretty","SDDS",
} ;

static char *USAGE1 = "cavget [-list=<string>[=<value>][,<string>[=<value>]...]]\n\
[-range=begin=<integer>,end=<integer>[,format=<string>][,interval=<integer>]]\n\
[-floatformat=<printfString>] \n\
{ [-delimiter=<string>] [-labeled] [-noQuotes] [-embrace=start=<string>,end=<string>] \n\
    | [-cavputForm]  }\n\
[-statistics=number=<value>,pause=<value>[,format=[tagvalue][pretty][SDDS,file=<filename]]] \n\
[-pendIoTime=<seconds>] [-repeat=number=<integer>,pause=<seconds>[,average[,sigma]]]\n\
[-numerical] [-errorValue=<string>] [-excludeErrors]\n\
[-despike[[neighbors=<integer>][,passes=<integer>][,averageOf=<integer>][,threshold=<value>]] \n\
[-printErrors] \n\n\
-list        specifies PV name string components\n\
-range       specifies range of integers and format string\n\
-floatFormat specifies printf-style format string for printing\n\
             single- and double-precision values.  %g is the default.\n\
-delimiter   specifies string to print between values.  The \n\
             default is a newline.\n\
-labeled     specifies that the PV name be echoed with the value.\n\
             A space separates the name and value.\n\
-noQuotes    Don't put quotes around string values that contain whitespace\n\
             or are blank.\n\
-embrace     Specifies putting braces around tag-value pairs.\n";
static char *USAGE2 = "             Typically used only with -label.\n\
-cavputForm  Specifies printout output in a format that can be given directly to\n\
             cavput for restoration of readback values.\n\
-pendIoTime  specifies maximum time to wait for connections and\n\
             return values.  Default is 1.0s \n\
-dryRun      specifies showing PV names and values but\n\
             not sending to IOCs.\n\
-errorValue  specifies a string to emit when there is a timeout or other\n\
             error for a PV.  The default is ?.\n\
-excludeErrors\n\
             specifies that when there is a timeout or other error for a PV,\n\
             there should be no output for that PV.\n\
-numerical   return a numerical value rather than a string for enumerated\n\
             types.\n\
-repeat      specifies repeated readings of the PVs at a specified time\n\
             interval.  If the average qualifier is given, then numerical\n\
             readbacks are averaged; for nonnumerical outputs, the last value\n\
             is given.\n\
-statistics  specifies statistical calculations of the PVs with reading times of \n\
             giving number at a specified time interval. \n\
             if format=tagvalue is given, the output format would be \n\
             \"mean <value> median <value> sigma <value> ....\" \n\
             if format=pretty, the statistical values are printed in columns format.\n\
             if format=SDDS,file=<filename> is given, the results will be written \n\
             to file. \n";
static char *USAGE3 = "-despike     despike must come along with repeat or statistics option which supplies\n\
             the number of readings. Despike specifies despiking of the number of \n\
             reading values given by repeat or statistics under the assumption\n\
             that consecutive readings are nearest neighbors. While, \n\
             <neighbors> is the number of neighbors for checking \n\
             despiking; <averageOf> is the number of consecutive readings around \n\
             the despiking for averaging and the averaged value (excluding the despike)\n\
             is going to replace this despiking data; <passes> is the number of times for\n\
             processing despiking; <threshold> is the despiking threshold, if the threshold\n\
             is zero or if the sum of absolute delta from the left most neighbor to \n\
             current reading is greater than <threshold>*<neighbors>, this reading value \n\
             is  going to be replace by the averaged value of <averageOf> consecutive\n\
             readings around this reading and excluding this reading, i.e., if threshold\n\
             is zero, despiking works similar to averaging. The value of <neighbors> and \n\
             <averageOf> must be smaller than the number of readings. \n\
printErrors  print error message to standard error if channel not found.\n\n\
Program by Michael Borland, ANL/APS ("EPICS_VERSION_STRING", "__DATE__")\n";

#include <sys/types.h>
#if defined(_WIN32)
#include <winsock.h>
#else
#include <sys/time.h>
#if defined(__APPLE__)
#include <sys/types.h>
#include <unistd.h>
#endif
#endif
void sleep_us(unsigned int nusecs) {
  struct timeval tval;

  tval.tv_sec = nusecs / 1000000;
  tval.tv_usec = nusecs % 1000000;
  select(0, NULL, NULL, NULL, &tval);
}

#define REPEAT_AVERAGE 0x0001U
#define REPEAT_SIGMA   0x0002U

long printResult(FILE *fp, char *startBrace, long labeled, long connectedInTime,
                 long channelType, PV_VALUE PVvalue, char *endBrace,
                 char *floatFormat, char *errorValue, long quotes, 
                 long cavputForm, long excludeErrors, 
                 long delimNeeded, char *delimiter,long secondValue, long doRepeats,
                 long doStats, long repeats, double *value, long average, short printErrors);
void printResultPretty (FILE *fp, long *channelType, PV_VALUE *PVvalue, long PVs,long quotes,
                        short *connectedInTime,char *errorValue, short printErrors);
long writeStatisticsToSDDSFile(char *outputFile,long *channelType, PV_VALUE *PVvalue,long PVs,
                               short *connectedInTime,char *errorValue, short printErrors);
void oag_ca_exception_handler (struct exception_handler_args args);

int main(int argc, char **argv)
{
  int32_t beginRange, endRange, rangeInterval;
  long dryRun;
  long PVs, j, i_arg, labeled, quotes, numerical, i;
  PV_VALUE *PVvalue;
  TERM_LIST *List;
  unsigned long flags;
  SCANNED_ARG *s_arg;
  char *rangeFormat, *floatFormat, *delimiter;
  double pendIOTime, repeatPause,despikeThreshold;
  chid *channelID;
  short *connectedInTime=NULL;
  char *startBrace, *endBrace;
  long average, sigma, repeats0, cavputForm, excludeErrors, delimNeeded,doStats,doRepeat,doDespike;
  int32_t repeats, despikeNeighbors,despikePasses,despikeAverageOf;
  char *errorValue = "?";
  char *statsFormat, *statsOutput;
  double **value=NULL;
  long *channelType=NULL;
  short printErrors = 0;
  
  argc = scanargs(&s_arg, argc, argv);
  if (argc<2) {
    fprintf(stderr,"%s%s%s\n", USAGE1, USAGE2, USAGE3);
    exit(1);
  }
  statsFormat=statsOutput=NULL;
  PVvalue = NULL;
  List = NULL;
  PVs = dryRun = labeled = numerical = 0;
  floatFormat = "%g";
  delimiter = "\n";
  quotes = 1;
  pendIOTime = 1.0;
  startBrace = endBrace = NULL;
  repeats = average = sigma = 0;
  doStats=doRepeat=doDespike=0;
  repeatPause = 1;
  cavputForm = excludeErrors = 0;
  channelID=NULL;
  despikeNeighbors=10;
  despikePasses=2;
  despikeAverageOf=4;
  despikeThreshold=0;
  
  for (i_arg=1; i_arg<argc; i_arg++) {
    if (s_arg[i_arg].arg_type==OPTION) {
      delete_chars(s_arg[i_arg].list[0], "_");
      switch (match_string(s_arg[i_arg].list[0], commandline_option, COMMANDLINE_OPTIONS, 0)) {
      case CLO_LIST:
        if (s_arg[i_arg].n_items<2)
          SDDS_Bomb("invalid -list syntax (cavget)");
        List = trealloc(List, sizeof(*List)*(s_arg[i_arg].n_items-1));
        for (j=1; j<s_arg[i_arg].n_items; j++) {
          List[j-1].flags = 0;
          List[j-1].string= s_arg[i_arg].list[j];
          List[j-1].value = NULL;
        }
        multiplyWithList(&PVvalue, &PVs, List, s_arg[i_arg].n_items-1);
        break;
      case CLO_RANGE:
        s_arg[i_arg].n_items --;
        rangeFormat = NULL;
        rangeInterval = 1;
        if (!scanItemList(&flags, s_arg[i_arg].list+1, &s_arg[i_arg].n_items, 0,
                          "begin", SDDS_LONG, &beginRange, 1, BEGIN_GIVEN,
                          "end", SDDS_LONG, &endRange, 1, END_GIVEN,
                          "interval", SDDS_LONG, &rangeInterval, 1, INTERVAL_GIVEN,
                          "format", SDDS_STRING, &rangeFormat, 1, FORMAT_GIVEN,
                          NULL) ||
            !(flags&BEGIN_GIVEN) || !(flags&END_GIVEN) || beginRange>endRange ||
            (flags&FORMAT_GIVEN && SDDS_StringIsBlank(rangeFormat)))
          SDDS_Bomb("invalid -range syntax/values");
        if (!rangeFormat)
          rangeFormat = "%ld";
        multiplyWithRange(&PVvalue, &PVs, beginRange, endRange, rangeInterval, rangeFormat);
        s_arg[i_arg].n_items ++;
        break;
      case CLO_PENDIOTIME:
        if (s_arg[i_arg].n_items!=2)
          SDDS_Bomb("wrong number of items for -pendIoTime");
        if (sscanf(s_arg[i_arg].list[1], "%lf", &pendIOTime)!=1 ||
            pendIOTime<=0)
          SDDS_Bomb("invalid -pendIoTime value (cavget)");
        break;
      case CLO_DRYRUN:
        dryRun = 1;
        break;
      case CLO_LABELED:
        labeled = 1;
        break;
      case CLO_PRINTERRORS:
        printErrors = 1;
        break;
      case CLO_FLOATFORMAT:
        if (s_arg[i_arg].n_items!=2)
          SDDS_Bomb("wrong number of items for -floatFormat");
        floatFormat = s_arg[i_arg].list[1];
        break;
      case CLO_DELIMITER:
        if (s_arg[i_arg].n_items!=2)
          SDDS_Bomb("wrong number of items for -delimiter");
        delimiter = s_arg[i_arg].list[1];
        break;
      case CLO_NOQUOTES:
        quotes = 0;
        break;
      case CLO_EMBRACE:
        s_arg[i_arg].n_items --;
        startBrace = "{";
        endBrace = "}";
        if (!scanItemList(&flags, s_arg[i_arg].list+1, &s_arg[i_arg].n_items, 0,
                          "start", SDDS_STRING, &startBrace, 1, 0,
                          "end", SDDS_STRING, &endBrace, 1, 0,
                          NULL))
          SDDS_Bomb("invalid -embrace syntax");
        s_arg[i_arg].n_items ++;
        break;
      case CLO_DESPIKE:
        s_arg[i_arg].n_items--;
        doDespike=1;
        if (!scanItemList(&flags, s_arg[i_arg].list+1, &s_arg[i_arg].n_items, 0,
                          "neighbors",SDDS_LONG, &despikeNeighbors, 1, 0,
                          "passes", SDDS_LONG, &despikePasses, 1, 0,
                          "averageOf", SDDS_LONG, &despikeAverageOf, 1, 0,
                          "threshold", SDDS_DOUBLE, &despikeThreshold, 1, 0,
                          NULL) || despikeNeighbors<0 || 
            despikePasses<0 || despikeAverageOf<0 || despikeThreshold<0 )
          SDDS_Bomb("invalid -despike syntax");
        s_arg[i_arg].n_items++;
        break;
      case CLO_REPEAT:
        s_arg[i_arg].n_items--;
        doRepeat=1;
        repeats = average = 0;
        if (!scanItemList(&flags, s_arg[i_arg].list+1, &s_arg[i_arg].n_items, 0,
                          "number", SDDS_LONG, &repeats, 1, 0,
                          "pause", SDDS_DOUBLE, &repeatPause, 1, 0,
                          "average", -1, NULL, 0, REPEAT_AVERAGE,
                          "sigma", -1, NULL, 0, REPEAT_SIGMA,
                          NULL) || repeats<1)
          SDDS_Bomb("invalid -repeat syntax");
        if (flags&REPEAT_AVERAGE)
          average = 1;
        if (flags&REPEAT_SIGMA)
          sigma = 1;
        s_arg[i_arg].n_items++;
        break;
      case CLO_STATISTICS:
        s_arg[i_arg].n_items--;
        doStats=1;
        repeats = 0;
        if (!scanItemList(&flags, s_arg[i_arg].list+1, &s_arg[i_arg].n_items, 0,
                          "number", SDDS_LONG, &repeats, 1, 0,
                          "pause", SDDS_DOUBLE, &repeatPause, 1, 0,
                          "format", SDDS_STRING, &statsFormat, 1,0,
                          "file", SDDS_STRING,&statsOutput, 1,0,
                          NULL) || repeats<1)
          SDDS_Bomb("invalid -statistics syntax");
        if (!statsFormat)
          SDDS_Bomb("invalid -statistics syntax, the format is not given.");
        switch (match_string(statsFormat, statisticsFormat, STATISICS_FORMATS, 0)) {
        case CLO_STATS_TAGVALUE:
        case CLO_STATS_PRETTY:
          if (statsOutput)
            SDDS_Bomb("invalid -statics format syntax");
          break;
        case CLO_STATS_SDDS:
          if (!statsOutput)
            SDDS_Bomb("invalid -statics format=SDDS syntax, output filename is not given");
          break;
        default:
          SDDS_Bomb("unknown format given to -statistics option");
          break;
        }
        s_arg[i_arg].n_items++;
        break; 
      case CLO_ERRORVALUE:
        if (s_arg[i_arg].n_items!=2)
          SDDS_Bomb("invalid -errorValue syntax");
        errorValue = s_arg[i_arg].list[1];
        break;
      case CLO_NUMERICAL:
        numerical = 1;
        break;
      case CLO_CAVPUTFORM:
        cavputForm = 1;
        break;
      case CLO_EXCLUDEERRORS:
        excludeErrors = 1;
        break;
      default:
        SDDS_Bomb("unknown option (cavget)");
        break;
      }
    }
    else
      SDDS_Bomb("unknown option");
  }
  
  if (doRepeat && doStats)
    SDDS_Bomb("-repeat and -statistics are imcompatible.");
  
  /*if (cavputForm && doStats)
    SDDS_Bomb("-cavputForm and -statistics are imcompatible."); */
  if (doDespike) {
    if (!doRepeat && !doStats)
      SDDS_Bomb("neither repeat or statistics is given, can not do despiking!");
    if (doRepeat) {
      sigma=0;
      average=1;
    }
    if (despikeNeighbors>repeats)
      despikeNeighbors=repeats;
    if (despikeAverageOf>repeats)
      despikeAverageOf=repeats;
  }
  if (cavputForm) 
    delimiter = ",";
  if (dryRun) {
    for (j=0; j<PVs; j++)
      printf("%32s\n", PVvalue[j].name);
  }
  else {
    ca_task_initialize();
    ca_add_exception_event ( oag_ca_exception_handler, NULL );
    if (!(channelID = (chid*)malloc(sizeof(*channelID)*PVs)) ||
        !(channelType = (long*)malloc(sizeof(*channelType)*PVs)) ||
        !(connectedInTime = (short*)malloc(sizeof(*connectedInTime)*PVs)))
      SDDS_Bomb("memory allocation failure (cavget)");
    for (j=0; j<PVs; j++)
      if (ca_search(PVvalue[j].name, channelID+j)!=ECA_NORMAL) {
        fprintf(stderr, "error (cavget): problem doing search for %s\n", 
                PVvalue[j].name);
	ca_task_exit();
        exit(1);
      }
    ca_pend_io(pendIOTime);
    if (numerical)
      for (j=0; j<PVs; j++) {
        channelType[j] = ca_field_type(channelID[j]);
        if ((channelType[j] = ca_field_type(channelID[j]))==DBF_ENUM)
          channelType[j] = DBF_DOUBLE;
      }
    else
      for (j=0; j<PVs; j++) 
        channelType[j] = ca_field_type(channelID[j]);
    if (!(value=malloc(sizeof(*value)*PVs)))
      SDDS_Bomb("memory allocation failure (cavget)");
    for (j=0;j<PVs;j++) {
      if (repeats) {
        if (!(value[j]=(double*)calloc((repeats), sizeof(**value))))
          SDDS_Bomb("memory allocation failure (cavget)");
      }
      else 
        if (!(value[j]=(double*)calloc((repeats+1), sizeof(**value))))
          SDDS_Bomb("memory allocation failure (cavget)");
    }
    repeats0 = 0;
    do {
      for (j=0; j<PVs; j++) {
        if (!(connectedInTime[j] = ca_state(channelID[j])==cs_conn))
          continue;
        switch (channelType[j]) {
        case DBF_STRING: 
        case DBF_ENUM:
          PVvalue[j].value = malloc(sizeof(char)*256);
          if (ca_get(DBR_STRING, channelID[j], PVvalue[j].value)
              !=ECA_NORMAL) {
            fprintf(stderr, "problem doing get for %s\n",
                    PVvalue[j].name);
	    ca_task_exit();
            exit(1);
          }
          break;
        case DBF_NO_ACCESS:
          fprintf(stderr, "Error: No access to PV %s\n", PVvalue[j].name);
	  ca_task_exit();
          exit(1);
          break;
        default:
          if (ca_get(DBR_DOUBLE, channelID[j], 
                     &value[j][repeats0])!=ECA_NORMAL) {
            fprintf(stderr, "problem doing get for %s\n",
                    PVvalue[j].name);
	    ca_task_exit();
            exit(1);
          }
          break;
        }
      }
      ca_pend_io(pendIOTime);
      repeats0++;
      if (!repeats) {
        for (j=0; j<PVs; j++) 
          PVvalue[j].mean=value[j][0];
      }
      if ((repeats0< repeats) && (repeatPause>0)) {
        ca_pend_event(repeatPause);
      }
    } while (repeats0 < repeats);
  
    if (repeats) {
      for (j=0;j<PVs;j++) {
        if (doDespike) {
          despikeData(value[j],repeats,despikeNeighbors,despikePasses,despikeAverageOf,
                      despikeThreshold,0);
        }
        find_min_max(&PVvalue[j].min,&PVvalue[j].max,value[j],repeats);
        computeMoments(&PVvalue[j].mean,&PVvalue[j].rms,&PVvalue[j].stDev,&PVvalue[j].MAD,
                       value[j],repeats);
        PVvalue[j].spread=PVvalue[j].max-PVvalue[j].min;
        PVvalue[j].sigma=PVvalue[j].stDev/sqrt(repeats);
        find_median(&PVvalue[j].median,value[j],repeats);
      }
    }
  
    delimNeeded = 0;
    if (!statsFormat || 
        match_string(statsFormat, statisticsFormat, STATISICS_FORMATS, 0)==CLO_STATS_TAGVALUE) {
      for (j=0; j<PVs; j++) {
        if (printResult(stdout, startBrace, labeled, connectedInTime[j], 
                        channelType[j], PVvalue[j], endBrace,
                        floatFormat, errorValue, quotes, cavputForm,
                        excludeErrors, delimNeeded, delimiter,sigma,doRepeat,doStats,repeats,value[j],average, 
                        printErrors))
          delimNeeded = 1;
      }
      if (delimNeeded)
        fputc('\n', stdout);
    } 
    if (statsFormat) {
      switch (match_string(statsFormat, statisticsFormat, STATISICS_FORMATS, 0)) {
      case CLO_STATS_PRETTY:
        printResultPretty (stdout, channelType, PVvalue, PVs, quotes,connectedInTime,errorValue, printErrors);
        break;
      case CLO_STATS_SDDS:
        writeStatisticsToSDDSFile(statsOutput,channelType, PVvalue,PVs,connectedInTime,errorValue, printErrors);
        break;     
      default:
        break;
      }
    }
    ca_task_exit();
  }
  for (j=0;j<PVs;j++) {
    if (PVvalue[j].name) free(PVvalue[j].name);
    if (PVvalue[j].value) free(PVvalue[j].value);
    PVvalue[j].name=PVvalue[j].value=NULL;
  }
  if (value) {
    for (j=0;j<PVs;j++)
      free(value[j]);
    free(value);
  }
  if (PVvalue)
    free(PVvalue);
  if (channelID)
    free(channelID);
  if (channelType)
    free(channelType);
  if (connectedInTime)
    free(connectedInTime);
  free_scanargs(&s_arg,argc);
  if (List)
    free(List);
  
  return 0;
}

long printResult(FILE *fp, char *startBrace, long labeled, long connectedInTime,
                 long channelType, PV_VALUE PVvalue, char *endBrace,
                 char *floatFormat, char *errorValue, long quotes, 
                 long cavputForm, long excludeErrors,
                 long delimNeeded, char *delimiter, long secondValue, 
                 long doRepeat, long doStats, long repeats, double *value, long average, short printErrors)
{
  long i;
  
  if (excludeErrors && !connectedInTime)
    return 0;
  if (delimNeeded && delimiter)
    fputs(delimiter, fp);
  if (cavputForm) {
    fprintf(fp, "%s=", PVvalue.name);
    quotes = 0;
  } else {
    if (startBrace)
      fputs(startBrace, fp);
    if (labeled)
      fprintf(fp, "%s ", PVvalue.name);
  }
  if (!connectedInTime) {
    if (printErrors)
      fprintf(stderr, "%s is not connected or invalid PV name.", PVvalue.name);
    else
      fputs(errorValue, fp);
  } else {
    switch (channelType) {
    case DBF_SHORT: case DBF_LONG: case DBF_CHAR: 
      if (cavputForm || (!doRepeat && !doStats)) {
        fprintf(fp, "%.0f", PVvalue.mean);
      } else {
        if (doRepeat) {
          if (!average) {
            for (i=0;i<repeats;i++) {
              fprintf(fp, "%.0f ", value[i]);
            }
          } else {
            fprintf(fp, "%.0f", PVvalue.mean);
            if (secondValue) {
              fprintf(fp, " ");
              fprintf(fp, "%.0f", PVvalue.sigma);
            }
          }
        }
        if (doStats) {
          fprintf(fp, "mean ");
          fprintf(fp, "%.0f",PVvalue.mean);
          fprintf(fp, " median ");
          fprintf(fp, "%.0f",PVvalue.median);
          fprintf(fp, " sigma ");
          fprintf(fp, "%.0f",PVvalue.sigma);
          fprintf(fp, " stDev ");
          fprintf(fp, "%.0f",PVvalue.stDev);
          fprintf(fp, " MAD ");
          fprintf(fp, "%.0f",PVvalue.MAD);
          fprintf(fp, " min ");
          fprintf(fp, "%.0f",PVvalue.min);
          fprintf(fp, " max ");
          fprintf(fp, "%.0f",PVvalue.max);
          fprintf(fp, " spread ");
          fprintf(fp, "%.0f",PVvalue.spread);
        }
      }
      break;
    case DBF_STRING: case DBF_ENUM:
      if (!quotes)
        fputs(PVvalue.value, fp);
      else {
        if (SDDS_StringIsBlank(PVvalue.value) ||
            SDDS_HasWhitespace(PVvalue.value))
          fprintf(fp, "\"%s\"", PVvalue.value);
        else
          fputs(PVvalue.value, fp);
      }
      break;
    case DBF_DOUBLE: case DBF_FLOAT:  
    default:
      if (cavputForm || (!doRepeat && !doStats)) {
        fprintf(fp, floatFormat, PVvalue.mean);
      } else {
        if (doRepeat) {
          if (!average) {
            for (i=0;i<repeats;i++) {
              fprintf(fp, floatFormat, value[i]);
              fprintf(fp," ");
            }
          } else {
            fprintf(fp, floatFormat, PVvalue.mean);
            if (secondValue) {
              fprintf(fp, " ");
              fprintf(fp, floatFormat, PVvalue.sigma);
            }
          }
        }
        if (doStats) {
          fprintf(fp, " mean ");
          fprintf(fp, floatFormat,PVvalue.mean);
          fprintf(fp, " median ");
          fprintf(fp, floatFormat,PVvalue.median);
          fprintf(fp, " sigma ");
          fprintf(fp, floatFormat,PVvalue.sigma);
          fprintf(fp, " stDev ");
          fprintf(fp, floatFormat,PVvalue.stDev);
          fprintf(fp, " MAD ");
          fprintf(fp, floatFormat,PVvalue.MAD);
          fprintf(fp, " min ");
          fprintf(fp, floatFormat,PVvalue.min);
          fprintf(fp, " max ");
          fprintf(fp, floatFormat,PVvalue.max);
          fprintf(fp, " spread ");
          fprintf(fp, floatFormat,PVvalue.spread);
        }
      }
      break;
    }
  }
  if (endBrace && !cavputForm)
    fputs(endBrace, fp);
  return 1;
}

void printResultPretty (FILE *fp, long *channelType, PV_VALUE *PVvalue, long PVs, long quotes,
                        short *connectedInTime,char *errorValue, short printErrors) {
  long i;
 
  fprintf(fp, "%20s","ControlName");
  fprintf(fp, "%10s","mean");
  fprintf(fp, "%10s","median");
  fprintf(fp, "%10s","sigma");
  fprintf(fp, "%10s","stDev");
  fprintf(fp, "%10s","MAD");
  fprintf(fp, "%10s","min");
  fprintf(fp, "%10s","max");
  fprintf(fp, "%10s","spread");
  fputc('\n', fp);
  for (i=0;i<100;i++)
    fputc('-',fp);
  fputc('\n',fp);
  
  for (i=0;i<PVs;i++) {
    fprintf(fp,"%20s",PVvalue[i].name);
    if (!connectedInTime[i]) {
      if (printErrors)
        fprintf(stderr, "%s is not connected or invalid PV name.", PVvalue[i].name); 
      else {
        fprintf(fp,"      ");
        fputs(errorValue, fp); 
      }
    }
    else {
      switch (channelType[i]) {
      case DBF_STRING: case DBF_ENUM:
        if (!quotes)
          fputs(PVvalue[i].value, fp);
        else {
          if (SDDS_StringIsBlank(PVvalue[i].value) ||
              SDDS_HasWhitespace(PVvalue[i].value))
            fprintf(fp, "\"%s\"", PVvalue[i].value);
          else
            fputs(PVvalue[i].value, fp);
        }
        break;
      default:
        fprintf(fp,"%10.4f%10.4f%10.4f%10.4f%10.4f%10.4f%10.4f%10.4f",PVvalue[i].mean,
                PVvalue[i].median,PVvalue[i].sigma,PVvalue[i].stDev,PVvalue[i].MAD,
                PVvalue[i].min,PVvalue[i].max,PVvalue[i].spread);
        break;
      }
    }
    fputc('\n',fp);
  }
}

long writeStatisticsToSDDSFile(char *outputFile,long *channelType, PV_VALUE *PVvalue,long PVs,
                               short *connectedInTime,char *errorValue, short printErrors)
{
  SDDS_DATASET SDDS_out;
  long i,row;
  
  for (i=0;i<PVs;i++) {
    switch (channelType[i]) {
    case DBF_STRING: 
    case DBF_ENUM:
    case DBF_CHAR:
      fprintf(stderr,"Error, do statistical analysis for PVs of non-numeric type is meaningless.");
      return 0;
    default:
      break;
    }
  }
  
  if(!SDDS_InitializeOutput(&SDDS_out, SDDS_BINARY, 0, NULL, "cavget output", outputFile)) 
    SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors|SDDS_EXIT_PrintErrors);
  if (SDDS_DefineColumn(&SDDS_out, "ControlName",NULL, NULL, NULL, NULL, SDDS_STRING, 0)<0 ||
      SDDS_DefineColumn(&SDDS_out, "mean",NULL, NULL, NULL, NULL, SDDS_DOUBLE, 0)<0 ||
      SDDS_DefineColumn(&SDDS_out, "median",NULL, NULL, NULL, NULL, SDDS_DOUBLE, 0)<0 ||
      SDDS_DefineColumn(&SDDS_out, "sigma",NULL, NULL, NULL, NULL, SDDS_DOUBLE, 0)<0 ||
      SDDS_DefineColumn(&SDDS_out, "stDev",NULL, NULL, NULL, NULL, SDDS_DOUBLE, 0)<0 ||
      SDDS_DefineColumn(&SDDS_out, "MAD",NULL, NULL, NULL, NULL, SDDS_DOUBLE, 0)<0 ||
      SDDS_DefineColumn(&SDDS_out, "minimum",NULL, NULL, NULL, NULL, SDDS_DOUBLE, 0)<0 ||
      SDDS_DefineColumn(&SDDS_out, "maximum",NULL, NULL, NULL, NULL, SDDS_DOUBLE, 0)<0 ||
      SDDS_DefineColumn(&SDDS_out, "spread",NULL, NULL, NULL, NULL, SDDS_DOUBLE, 0)<0 )
    SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors|SDDS_EXIT_PrintErrors);
  if (!SDDS_WriteLayout(&SDDS_out))
    SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors|SDDS_EXIT_PrintErrors);
  if (!SDDS_StartTable(&SDDS_out, PVs)) 
    SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors|SDDS_EXIT_PrintErrors);
  for (i=row=0;i<PVs;i++,row++) {
    if (printErrors && !connectedInTime[i])
      fprintf(stderr, "%s is not connected or invalid PV name.", PVvalue[i].name);
    if (!connectedInTime[i] ||
        channelType[i]==DBF_STRING || channelType[i]==DBF_ENUM ) {
      row--;
      continue;
    } else {
      if (!SDDS_SetRowValues(&SDDS_out,SDDS_SET_BY_NAME|SDDS_PASS_BY_VALUE,row,
                             "ControlName",PVvalue[i].name,
                             "mean",PVvalue[i].mean,"median",PVvalue[i].median,
                             "sigma",PVvalue[i].sigma,"stDev",PVvalue[i].stDev,
                             "MAD",PVvalue[i].MAD,"minimum",PVvalue[i].min,
                             "maximum",PVvalue[i].max,"spread",PVvalue[i].spread,NULL)) {
        SDDS_SetError("Unable to set statistical values in SDDS data set");
        SDDS_PrintErrors(stderr, SDDS_EXIT_PrintErrors|SDDS_VERBOSE_PrintErrors);
        return 0;
      }
    }
  }
  if (!SDDS_UpdatePage(&SDDS_out,0))
    SDDS_PrintErrors(stderr,SDDS_EXIT_PrintErrors|SDDS_VERBOSE_PrintErrors);
  if (!SDDS_Terminate(&SDDS_out))
    SDDS_PrintErrors(stderr,SDDS_EXIT_PrintErrors|SDDS_VERBOSE_PrintErrors);
  return 1;
}

void oag_ca_exception_handler (struct exception_handler_args args) {
  char *pName;
  int severityInt;
  static const char *severity[] = 
    {
      "Warning",
      "Success",
      "Error",
      "Info",
      "Fatal",
      "Fatal",
      "Fatal",
      "Fatal"
    };
  
  severityInt = CA_EXTRACT_SEVERITY(args.stat);
  
  fprintf(stderr, "CA.Client.Exception................\n");
  fprintf(stderr, "    %s: \"%s\"\n",
          severity[severityInt], 
          ca_message ( args.stat ) );
  
  if (args.ctx) {
    fprintf(stderr, "    Context: \"%s\"\n",args.ctx);
  }
  if ( args.chid ) {
    pName = (char*)ca_name ( args.chid );
    fprintf(stderr, "    Channel: \"%s\"\n", pName);
    fprintf(stderr, "    Type: \"%s\"\n", dbr_type_to_text(args.type));
  }
  fprintf(stderr, "This sometimes indicates an IOC that is hung up.\n");
  _exit(1);
}
